import platform
if platform.system() == "Windows":
    import keyboard
else:
    from winhotkey.dummy_keyboard import keyboard # For testing on non-windows systems
from time import sleep

shift_list = '~!@#$%^&*()_+{}|:"<>?QWERTYUIOPLKJHGFDSAZXCVBNM'

def sender(phrase, delay=3.0):
    """
    Will type the phrase by breaking phrase into indiv characters and using keyboard.send()
    """
    sleep(delay)
    for c in phrase:
        if c in shift_list:
            # Since using .send() I have to detect characters that are on the "shift" side of the key and send the shift with it
            # Assumes caps lock is off
            c = "shift+" + c.lower()
        keyboard.send(c)


"""
This file is a dummy fill in for the actual keyboard module
Mainly because I develop mostly on Mac but keyboard generally only works right on Windows
So this provides feedback for functions being called, but does not actually work
"""
from time import sleep

class keyboard:

    @classmethod
    def add_hotkey(cls, key_to_add, fxn, args):
        print(f"Would have activated {key_to_add} with {args}")

    @classmethod
    def remove_hotkey(self, key_to_remove):
        print(f"Would disconnect {key_to_remove}")

    @classmethod
    def write(self, phrase):
        print(phrase)

    @classmethod
    def wait(self, key):
        print(f"Would wait until {key} pressed.  Will need Ctrl-C to quit")
        while True:
            sleep(0.05)

    @classmethod
    def send(self, key):
        print(f"Would send {key}")


"""
Taken from: https://github.com/zauberzeug/nicegui/issues/1956
Discussed at: https://github.com/encode/uvicorn/discussions/1103
"""

import asyncio
import threading
import uvicorn
from uvicorn.config import Config

class ThreadedUvicorn:
    def __init__(self, config: Config):
        self.server = uvicorn.Server(config)
        self.thread = threading.Thread(daemon=True, target=self.server.run)

    def start(self):
        self.thread.start()
        asyncio.run(self.wait_for_started())

    async def wait_for_started(self):
        while not self.server.started:
            await asyncio.sleep(0.1)

    def stop(self):
        if self.thread.is_alive():
            self.server.should_exit = True
            while self.thread.is_alive():
                continue




import keyboard

def sendit(text):
    keyboard.write(text)


keyboard.add_hotkey('ctrl+shift+0', print, args=('triggered', 'hotkey'))

# Press PAGE UP then PAGE DOWN to type "foobar".
# keyboard.add_hotkey('ctrl+alt+9', lambda: keyboard.write('foobar'))
keyboard.add_hotkey('ctrl+windows+9', keyboard.write, args=('foobar',))
# keyboard.add_hotkey('ctrl+alt+9', sendit, args=('foobar',))

print("Press esc")

# Blocks until you press esc.
keyboard.wait('esc')


from fastapi import FastAPI, Query, HTTPException, Response
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse
from pydantic import BaseModel, Field 
import platform
if platform.system() == "Windows":
    import keyboard
else:
    from winhotkey.dummy_keyboard import keyboard # For testing on non-windows systems
from time import sleep
from typing import Annotated, List
from importlib import resources
# import asyncio
import logging

from winhotkey.keyboard_wrappers import sender

api_app = FastAPI()
logger = logging.getLogger(__name__)

static_path = None
with resources.path('winhotkey', 'web') as web_path:
    static_path = web_path/"static"

api_app.mount(f"/static", StaticFiles(directory=static_path), name="static")

class HotKey(BaseModel):
    index: int = Field(description="Order this hotkey should be displayed on the gui", default=-1)
    assigned_key: str = Field(description="'keyboard' compatible hot key description string", min_length=1)
    phrase: str = Field(description="Phrase to type when hot key is pressed", default="")

current_hotkeys = {}
type_delay = 2.0
def initializeSettings(
        hotkey_prefix,
        delay_in_seconds
    ):
    global current_hotkeys
    global type_delay
    possible_entries = [f"{i}" for i in range(1,10)]
    possible_entries.append("0")
    possible_entries = possible_entries[::-1]
    for index, num in enumerate(possible_entries):
        hkey = f"{hotkey_prefix}+{num}"
        current_hotkeys[index] = HotKey(index=index, assigned_key=hkey, phrase="")
    
    type_delay = delay_in_seconds

@api_app.get("/")
def home():
    """
    Redirect to the main page
    """
    tosend = """
        <!DOCTYPE html>
        <html>
            <head>
                <meta http-equiv="refresh" content="1; url='./static/pages/home.html'" />
            </head>
            <body>
                <p>Redirecting to main menu</p>
            </body>
        </html>
    """
    return HTMLResponse(tosend)

@api_app.get("/api/hotkeys")
def get_hotkeys() -> List[HotKey]:
    """
    Returns list of current hotkeys in display order.
    """
    hkeys = list(current_hotkeys.values())
    hkeys = sorted(hkeys, key=lambda hkey: hkey.index)
    return hkeys

@api_app.post("/api/hotkeys")
def set_hotkey(hot_key: HotKey):
    """
    Will assign (or reassign) hot key
    """
    existing_hotkey = current_hotkeys.get(hot_key.index, None)
    if existing_hotkey is not None:
        # Remove the existing assignment (catching if not assigned)
        try:
            keyboard.remove_hotkey(existing_hotkey.assigned_key)
        except KeyError:
            # Wasn't there, no harm
            pass
        except Exception as badnews:
            print(f"Couldn't remove {existing_hotkey.assigned_key} because {badnews}")

    # attach to hot keys
    keyboard.add_hotkey(hot_key.assigned_key, sender, args=(hot_key.phrase, 0.5, ))
    # Store in Dict
    current_hotkeys[hot_key.index] = hot_key
    
    return {}


@api_app.post("/api/hotkeys")
def set_hotkey(hot_key: HotKey):
    """
    Will assign (or reassign) hot key
    """
    # Check for key already assigned
    phrase = current_hotkeys.get(hot_key.assigned_key, None)
    if phrase is not None:
        # unassign
        keyboard.remove_hotkey(hot_key.assigned_key)
        
    # Assign hotkey
    keyboard.add_hotkey(hot_key.assigned_key, sender, args=(hot_key.phrase, 0.5, ))
    current_hotkeys[hot_key.assigned_key] = hot_key
    return {}

@api_app.get("/api/type_phrase")
def type_phrase(index: int = Query(description="Index of assigned key to activate"), 
                delay:float = Query(default=type_delay, description="The time in seconds we will wait before we begin typing")
):
    """
    Will type the assigned key after the delay in seconds.  This is to allow you to put focus on the correct window if the hot keys are not working in that window.
    """
    hkey = current_hotkeys.get(index, None)
    if hkey is not None:
        if len(hkey.phrase) > 0:
            sender(phrase=hkey.phrase, delay=delay)
        else:
            logger.warning(f"No phrase to type for hot key {hkey.assigned_key}")
    
    return {}




import keyboard
from time import sleep

shift_list = '~!@#$%^&*()_+{}|:"<>?QWERTYUIOPLKJHGFDSAZXCVBNM'

def sender(phrase, delay=3.0):
    sleep(delay)
    for c in phrase:
        if c in shift_list:
            c = "shift+" + c.lower()
        print(f"sending {c}")
        keyboard.send(c)






# SPDX-FileCopyrightText: 2023-present richmr <richmr@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
from enum import Enum
import typer
from typing_extensions import Annotated
import platform
if platform.system() == "Windows":
    import keyboard
else:
    from winhotkey.dummy_keyboard import keyboard # For testing on non-windows systems
from uvicorn.config import Config
from time import sleep
import logging

from winhotkey.__about__ import __version__
from winhotkey.ThreadedUvicorn import ThreadedUvicorn
from winhotkey.keyboard_wrappers import sender

cli_app = typer.Typer()


@cli_app.command()
def cli(hotkey_prefix: Annotated[str, typer.Option(help="'keyboard' compatible hot key prefix'")] = "shift+ctrl+alt",
        reverse_hotkey_order: Annotated[bool, typer.Option(help="Start with 1 and go to 0 or 0 and go to 1.  I actually prefer 0 to 1")] = True,
        hide_input: Annotated[bool, typer.Option(help="In case your 'phrases' are sensitive, you will not see what you enter")] = True):
    """
    Allows you store up to 10 'phrases' for instant recall by typing 'Shift+Ctrl+Alt[1-0]'\n
    Just hit enter to stop storing 'phrases'\n
    """
    possible_entries = [f"{i}" for i in range(1,10)]
    possible_entries.append("0")
    if reverse_hotkey_order:
        possible_entries = possible_entries[::-1]
    keys_to_add = {}
    for entry in possible_entries:
        key = f'{hotkey_prefix}+{entry}'
        phrase = typer.prompt(f"Please enter phrase for '{key}'", hide_input=hide_input, default="Press Enter when done")
        if phrase == "Press Enter when done":
            break
        keys_to_add[key] = phrase

    for (n, (key_to_add, phrase)) in enumerate(keys_to_add.items()):
        # Convert the phrase to string of characters separated by comma
        # keyboard.add_hotkey(key_to_add, keyboard.write, args=(phrase,))
        keyboard.add_hotkey(key_to_add, sender, args=(phrase,1.0,))
    
    print("Press SHIFT+CTRL+ALT+ESC to exit.  Leave this window open!")

    # Blocks until you press esc.
    keyboard.wait('shift+ctrl+alt+esc')

class LogLevels(str, Enum):
    CRITICAL = "critical"
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    DEBUG = "debug"

@cli_app.command()
def web(hotkey_prefix: Annotated[str, typer.Option(help="'keyboard' compatible hot key prefix'")] = "shift+ctrl+alt",
        type_delay: Annotated[float, typer.Option(help="Set the delay time in seconds when using the delayed typing feature")] = 2.0,
        logging_level: Annotated[LogLevels, typer.Option(help="Set the log level for the web server")] = LogLevels.WARNING.value):
    """
    Starts a web interface to configure winhotkey
    """
    from winhotkey.web import initializeSettings
    initializeSettings(hotkey_prefix, type_delay)

    config = Config("winhotkey.web:api_app", host="127.0.0.1", port=17455, reload=True, log_level=logging_level.value)
    server = ThreadedUvicorn(config)
    server.start()
    if logging_level not in [LogLevels.INFO, LogLevels.DEBUG]:
        print("Access the GUI at: http://127.0.0.1:17455/")
    print("Press SHIFT+ALT+CTRL+C to exit.  Leave this window open!")
    keyboard.wait('shift+alt+ctrl+c')
    server.stop()

def winhotkey():
    if not platform.system() == "Windows":
        print("Winhotkey only fully works on Windows.  It will be started in a test mode.  No hotkeys will actually work.")
    cli_app()




# SPDX-FileCopyrightText: 2023-present richmr <richmr@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
import sys

if __name__ == "__main__":
    from winhotkey.cli import winhotkey

    sys.exit(winhotkey())
